#!/usr/bin/pypy-1.9

import time
import sys

def fileLen(fname):
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
    return i + 1

def init(buffer):
   from glob import globs, setGlobals
   from array import array

   # create spatial variables of specified number
   temp = buffer.pop(0)
   Vars = int(temp.split()[0].strip())+1
   TypeId = temp.split()[1].strip()
   
   # initialize constraints
   N = fileLen(globs['calc']+"/"+globs['calc']+".relations")
   setGlobals("size", N)
   Id = 0
   from bitcoding import B_dict
   with open(globs['calc']+"/"+globs['calc']+".identity") as f:
      Id = B_dict[f.readline().strip()]
   if N <= 8:
      ConMatrix = tuple([array('B',[B_dict['DALL'] if i != j else Id for i in xrange(Vars)]) for j in xrange(Vars)])
      UConMatrix = tuple([array('B',[B_dict['DALL'] if i != j else Id for i in xrange(Vars)]) for j in xrange(Vars)])
   elif N <= 16:
      ConMatrix = tuple([array('H',[B_dict['DALL'] if i != j else Id for i in xrange(Vars)]) for j in xrange(Vars)])
      UConMatrix = tuple([array('H',[B_dict['DALL'] if i != j else Id for i in xrange(Vars)]) for j in xrange(Vars)])
   else:
      ConMatrix = tuple([array('I',[B_dict['DALL'] if i != j else Id for i in xrange(Vars)]) for j in xrange(Vars)])
      UConMatrix = tuple([array('I',[B_dict['DALL'] if i != j else Id for i in xrange(Vars)]) for j in xrange(Vars)])
  
   # parse spatial CSP
   from parsecsp import parsecsp
   parsecsp(ConMatrix, buffer)

   return TypeId, ConMatrix, UConMatrix

class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg

# main function
def main(argv=None):

   if argv is None:
        argv = sys.argv

   # set some sys useful stuff
   sys.setrecursionlimit(1<<30)
   sys.setcheckinterval(10000)

   from optparse import OptionParser

   # initialize parser for command line arguments
   parser = OptionParser()
   
   # set parsing options
   parser.add_option("-k", "--bound", type="int", dest="bound", default=0, help="interaction mode: 0, 1, 2, ... [default: %default]")
   parser.add_option("-c", "--calculus", type="string", dest="calc", default="rcc8", help="interaction mode: rcc8, or allen [default: %default]")
   parser.add_option("-m", "--method", type="string", dest="method", default="iterative", help="interaction mode: recursive, or iterative [default: %default]")
   parser.add_option("-p", "--pcheuristic", type="string", dest="pcheuristic", default="none", help="interaction mode: none, or weighted [default: %default]")
   parser.add_option("-s", "--split", type="string", dest="split", default="base", help="interaction mode: base, or horn [default: %default]")
   parser.add_option("-d", "--print", action="store_true", dest="printsol", help="print solution to output")
   parser.add_option("-z", "--pc", action="store_true", dest="pc", help="path consistency only")
   parser.add_option("-e", "--heuristic", type="string", dest="heuristic", default="global", help="interaction mode: global, or none [default: %default]")
   parser.add_option("-f", "--file", action="store", type="string", dest="filename")

   # parse command line arguments
   options, args = parser.parse_args()

   from glob import setGlobals
   #import argparse

   try:   

      # check and set globals for bound information
      if options.bound < 0:
         raise Usage("Invalid option: " + str(options.bound))

      # check and set globals for spliting information
      if options.split == "base":
         setGlobals('split',0)
      elif options.split == "horn":
         setGlobals('split',1)
      else:
         raise Usage("Invalid option: " + options.split)

      # check and set globals for calculus information
      if not (options.calc == "rcc8" or options.calc == "allen"):
         raise Usage("Invalid option: " + options.calc)
      else:
         setGlobals('calc',options.calc)

      # check and set globals for patch consistency heuristic information
      if options.pcheuristic == "none":
         setGlobals('pcheuristic',0)
      elif options.pcheuristic == "weighted":
         setGlobals('pcheuristic',1)
      else:
         raise Usage("Invalid option: " + options.pcheuristic)

      # check and set globals for consistency heuristic information
      if options.heuristic == "none":
         setGlobals('heuristic',0)
      elif options.heuristic == "global":
         setGlobals('heuristic',2)
      else:
         raise Usage("Invalid option: " + options.heuristic)

      # check and set globals for method decision information
      if options.method == "recursive":
         solver = "rec"    
      elif options.method == "iterative": 
         solver = "iter"
      else:
         raise Usage("Invalid option: " + options.method)

   except Usage, err:
        print err.msg
        parser.print_help()
        return 1

   with open(options.filename) as f:
      total = 0
      total_splits = 0
      meter = 0
      splittings = []
      buffer = []

      for i in f:
         buffer.append(i)

         if i.strip() == '.':
            start = elapsed = 0
            meter += 1

            from counters import counter

            TypeId, ConMatrix, UConMatrix = init(buffer)
            buffer = []

            neighbors = {}
          
            from glob import globs

            #if globs['pcheuristic'] == 0: # path consistency
            #   from spc import sPC as pc 
            #elif globs['pcheuristic'] == 1: # path consistency with weighted queue
            #   from spcw import sPCw as pc

            start = time.clock()

            solution = None
            #if pc(ConMatrix, neighbors):
            #   solution = ConMatrix
            #   if not options.pc:
            from consistency import consistency
            from iconsistency import iconsistency
            
            if solver == "rec": 
               solution, alpha = consistency(UConMatrix, ConMatrix, neighbors, len(ConMatrix)**2+1, options.bound, splittings)
            elif solver == "iter":
               solution, alpha = iconsistency(UConMatrix, ConMatrix, neighbors, len(ConMatrix)**2+1, options.bound, splittings)

            elapsed = time.clock() - start
            total += elapsed
            total_splits += sum(splittings)/float(len(splittings))

            # print values of complexity counters
            print '--------------------------------------'
            print 'Type ID: ', TypeId
            print '--------------------------------------'
            print 'Unsatisfied constraints: ', alpha
            print 'Visited nodes: ', counter['nodes']
            print 'Decisions: ', counter['decisions']
            print 'Splittings: ', sum(splittings)/float(len(splittings))
            #print 'Visited arcs: ', counter['arcs']
            #print 'Checked constraints: ', counter['cons']
            print '--------------------------------------'
            print 'Time: ', elapsed, 'sec'
            print '--------------------------------------'

            counter['arcs'] = 0
            counter['cons'] = 0
            counter['nodes'] = 0
            counter['decisions'] = 0
  
            # print solution  
            if solution != None:
               print 'Consistent'
               if options.printsol == True:
                  from printm import printMatrix
                  printMatrix(solution)
            else:
               print 'Inconsistent'
            print '--------------------------------------'

            ConMatrix = None
          
   print '======================================'
   print 'Total time:', total, 'sec'
   print 'Avg time per CSP:', total/(1.0*meter), 'sec'
   print 'Avg splits per CSP:', total_splits/(1.0*meter)
   print '======================================'

if __name__ == '__main__':
   sys.exit(main())

